#' @title k-places
#' @description Execute a structural equivalence analysis in bipartite graphs allowing a \emph{k} tolerance level
#' @usage kplaces(data, col.elements, col.sets, k)
#' @param data The input data frame. Data must be in a two-mode edgelist structure
#' @param col.elements The name of the column of elements
#' @param col.sets The name of the column of sets
#' @param k A natural number that indicates the tolerance level
#' @format The input format must be a data frame in a two-mode edgelist structure. \cr
#' Data frame can contain other variables, relevant columns will be selected in \emph{Arguments} parameters. Missing values, empty cells and repeated pairs (Element-Set) will be deleted for analysis.
#' @details A k-place is a place (a position of structural equivalence) which allows a maximum divergence of \emph{k} sets that define the place. Compared with \code{\link{places}}, \emph{k-places} are a less restrictive way of identifying positions of structural equivalence. This function identifies places and k-places, and a network of places and k-places is created.\cr
#' A network of k-places is defined as a collection of k-places and places linked by sharing sets. A pair of k-places and/or places are linked if they are defined by at least one set in common. A network of k-places, as a blockmodel image, is a network of structurally equivalent positions.\cr
#' The k-places will not be identified for cases with k sets or less.\cr
#' The ambiguous cases are reported as messages in the R console. An ambiguous case is a subset of more than two places where the dyads can be a k-place but not all the subset. For example, for k=1, the place P1 defined by the sets {A, B} can be a k-place with P2{A, B, C} or with P3{A, B, D}, but the k-place formed by the three places is not correct because P2 and P3 are divergent in more than k sets (specifically, the sets C and D)
#' @return The output is a list with four data frames:
#' @return \itemize{\item The original two columns data frame and the column "k_places" with k-places and places labels'
#' \item A data frame containing information about places and k-places.\cr
#' "PlaceLabel" contains places and k-places labels. Places labels start with \emph{P}, followed by the place number, the number of elements in place and the number of sets defining place. K-Places labels start with \emph{P}, followed by the k-place number, an \emph{*}, the number of elements in k-place, the number of all sets defining the k-place,the number of sets in common among the merged places, and the value of \emph{k} \cr
#' "NbElements" contains the number of elements in the place / k-place \cr
#' "NbSets" contains the number of sets defining the place / k-place \cr
#' "PlaceDetail" contains the name of all the elements in the place /k-place and all the sets defining the place / k-place
#' \item A data frame with the relation of places merged to k-places and the sets in common
#' \item The network of places in a two-mode edgelist format }
#' @author Delio Lucena Piquero (\href{http://lereps.sciencespo-toulouse.fr/}{LEREPS} - University of Toulouse) <delio.lucena@ut-capitole.fr>
#' @references The use of k-places is suggested in: \cr
#' Lucena, D. (2020). Beyond the Dyadic Approach in Social Network Analysis: Applications to Innovation Studies and Financial Economics (Doctoral dissertation).
#' @examples # Load example dataset
#' data(fooA)
#'
#' # Compute kplaces
#' Result <- kplaces(data = fooA, col.elements = "Actors", col.sets = "Groups", k = 1)
#' Result <- kplaces(fooA, "Actors", "Groups", 1)
#' @importFrom stats aggregate na.omit
#' @export


kplaces<-function(data, col.elements, col.sets, k) {
  message("Cases with ", k," or less sets will not be computed as k-places")
  #Subset data into two variable data frame, elements and sets
  pos.x<-which(colnames(data)==col.elements)
  pos.y<-which(colnames(data)==col.sets)
  k<-k
  DB<-data[,c(pos.x, pos.y)]
  #Remove empty cells and NAs. Report the information
  message("Cleaning data ... rows with empty cells and NAs will be removed")
  DB[DB==""]<-NA
  frm<-nrow(DB)
  DB<-na.omit(DB)
  message("Rows removed: ", c(frm - nrow(DB)))

  message("Cleaning data ... duplicate rows will be removed")
  frm<-nrow(DB)
  DB<-unique(DB)
  message("Duplicate rows removed: ", c(frm - nrow(DB)))

  #Identify the number of elements and sets
  message("There are ",length(unique(DB[,1]))," elements and ",length(unique(DB[,2]))," sets")

  message("Working ...")

  #First step is to identify the places
  #Rename column names to "Elem" and "Set"
  names(DB)[1]<-"Elem"
  names(DB)[2]<-"Set"

  #Identify the number of Sets by Element
  #Calculate the frequency of each Element in edgelist
  count_i<-as.data.frame(table(DB[,1]))
  #Add the result to data frame as new variable called "Freq"
  DB$count_i<-count_i[match(DB$Elem, count_i$Var1),"Freq"]

  #Create the variables "count_g", "count_dif_i", "ProtoPlaces" and "Places"
  DB$count_g<-NA
  DB$count_dif_i<-NA
  DB$ProtoPlaces<-NA
  DB$Places<-NA

  #Order the data by some variables
  DB<-DB[order(-DB[,3], DB[,1], DB[,2]),]

  #Create "count_i2". This object record all the differents number of Sets by place
  #Each value indicates that there is at least one place with that number of sets
  #This vector drives the loop
  count_i2<-unique(DB$count_i)

  for (i in count_i2) {
    subcase<-subset(DB, count_i==i)
    El<-unique(subcase[,1])
    #The values of the variable "Set" are aggregates as list
    subcase2<-aggregate(Set~Elem, data=subcase, FUN=list)
    #Information of variable "count_g" added
    subcase$count_g<-length(unlist(subcase2[1,2]))
    #Values of "Set" are reclassified in character to be aggregated again
    subcase2$Set<-sapply(subcase2$Set, paste, collapse=";")
    #Aggregate again in order to have all the elements together
    subcase3<-aggregate(Elem~Set, data=subcase2, FUN=list)
    #Count the number of elements per place
    subcase3$count_dif_i<-sapply(subcase3$Elem, function(x) length(unlist(x)))
    #The values of "Elem" are reclassified in character
    subcase3$Elem<-sapply(subcase3$Elem, paste, collapse=";")
    #Places are identified. Variables "Pelem" and "Pset" are added and "Protoplaces" is created
    subcase3$Pelem<-sapply(subcase3$Elem, paste, collapse=";")
    subcase3$Pset<-subcase3$Set
    subcase3$ProtoPlace<-paste0("{", subcase3$Pelem,"}" ," - ","{", subcase3$Pset,"}")
    #Variable "Pset" is created in subcase2
    subcase2$Pset<-subcase2$Set
    #Information is added to subcase. First the "Pset" from subcase2 and then "Pelem",
    #"count_dif_i" and "Protoplace" from subcase3 through "Pset"
    subcase<-merge(subcase, subcase2[,c("Elem","Pset")], by="Elem")
    subcase$count_dif_i<-NULL
    subcase<-merge(subcase, subcase3[,c("Pset","Pelem","count_dif_i" ,"ProtoPlace")], by="Pset")

    #Information about "Protoplaces" and "count_g" is added to DB
    for (j in El) {
      DB[,"ProtoPlaces"][which(DB$Elem == j)]<-subcase[,"ProtoPlace"][which(subcase$Elem == j)]
      DB[,"count_dif_i"][which(DB$Elem == j)]<-subcase[,"count_dif_i"][which(subcase$Elem == j)]
      DB[,"count_g"][which(DB$Elem == j)]<-subcase[,"count_g"][which(subcase$Elem == j)]
    }
  }

  #To add places' names a new loop is created for the suite
  #All places are listed in a new data frame. Places will be added to BD later.
  PP<-unique(DB$ProtoPlaces)
  PrPl<-as.data.frame(PP)
  PrPl$Cele<-NA
  PrPl$Cset<-NA

  for(i in PrPl$PP) {
    PrPl[,"Cele"][which(PrPl$PP==i)]<-unique(DB[,"count_dif_i"][which(DB$ProtoPlaces==i)])
    PrPl[,"Cset"][which(PrPl$PP==i)]<-unique(DB[,"count_g"][which(DB$ProtoPlaces==i)])
  }

  #Data frame is ordered by number of sets and by number of elements
  PrPl<-PrPl[order(-PrPl[,3], -PrPl[,2]),]

  #New variables are added and the label of each place is created
  PrPl$PNumber<-(1:length(PrPl$PP))
  PrPl$PName<-NA
  #For places' label we use the letter "P" followed by the number of the place and zeros on
  #the left, the number of elements and the number of sets
  for(i in PrPl$PP) {
    PrPl$PName<-paste0("P", formatC(PrPl$PNumber, width=nchar(max(PrPl$PNumber)), flag="0"),
                       "(",as.factor(PrPl$Cele),"-", PrPl$Cset,")")
  }

  #INformation of places is added to DB
  for(i in DB$ProtoPlaces) {
    DB[,"Places"][which(DB$ProtoPlaces==i)]<-PrPl[,"PName"][which(PrPl$PP==i)]
  }

  #Variable "count_g" is deleted because is the same than "count_i"
  DB$count_g<-NULL

  #We create the edgelist
  #DB is ordered by number of sets, number of elements and places
  DB<-DB[order(DB[,6]),]
  EdList <- unique(DB[, c(6, 2)]) #To identify kplaces ignore cases with less than k sets: unique(DB[which(DB$count_i>k), c(6, 2)])

  #The code for k-places identification starts here
  #We add a new variable on DB to store k-places, and another where store common sets.
  DB$k_places<-NA
  DB$Common_Sets<-NA
  DB[,8]<-sapply(DB[,8], as.character)

  #Create a place's matrix to identify the number of common sets by dyads
  bimod<-table(EdList[,1], EdList[,2])
  OrgMat<-bimod %*% t(bimod)
  diag(OrgMat)<-0
  OrgMat[lower.tri(OrgMat)]<-0
  #Transform the matrix on a edgelist
  #La siguiente linea sustituye a melt de reshape2
  OrgList<-cbind(expand.grid(dimnames(OrgMat)), value = as.vector(OrgMat))

  #Check if k is a valid value. If not, stop the function and print a message
  if (!(k %in% c(0:(max(DB[, 3])-1)))) {
    message("The highest number of sets in a place is ",max(DB[, 3]) ,". The argument k must be a natural number betweed 0 and ",
         (max(DB[, 3])-1))
    message("Stopping process")
    exit()
  }

  #Remove all dyads without shared sets
  OrgList<-OrgList[!(OrgList[,3]==0),]
  #Remove dyads which one of the places is defined by less than k sets
  overkplaces<-unique(DB[which(DB[,3]>k),6])
  OrgList<-subset(OrgList, OrgList[,1] %in% overkplaces & OrgList[,2] %in% overkplaces)

  #Check if there are dyads in OrgList
  if (length(OrgList[,1])==0) {
    message("Sorry. The value of k is too high. Try a lower value.")
    message("Stopping process")
    exit()
  }

  #Create variables Dif and Diff. Dif is the difference of sets in each dyad.
  #Diff is the number of non-common sets in each dyad. Is a measure of the similarity/difference beween places
  OrgList$Dif<-NA #Difference in the number of groups
  OrgList$Diff<-NA #Number of non shared groups
  for (i in c(1:nrow(OrgList))) {
    OrgList[i,4]<-abs((PrPl[,3][which(PrPl[,5] == OrgList[i,1])] - PrPl[,3][which(PrPl[,5] == OrgList[i,2])]))
    OrgList[i,5]<-abs((PrPl[,3][which(PrPl[,5] == OrgList[i,1])]-OrgList[i,3])+(PrPl[,3][which(PrPl[,5] == OrgList[i,2])]-OrgList[i,3]))
  }
  OrgList<-OrgList[(OrgList[,5] <= k),]
  OrgList<-OrgList[order(rev(OrgList[,3]), OrgList[,1], rev(OrgList[,4])),]

  #We create a dataframe with the values of lines existing in OrgList and the number of dyades
  #that takes these values. This dataframe determines the hierarchy of data analysis.
  OLCount<-as.data.frame(unique(OrgList[,3]))
  OLCount$Count<-NA
  OLCount<-OLCount[order(rev(OLCount[,1])),] #Start with those with more sets in common
  for(i in OLCount[,1]) {
    OLCount$Count[which(OLCount[,1]==i)]<-length(which(OrgList[,3]==i))
  }

  #Check that the cases of the subcase haven't appeared before, and those that don't contain
  #NAs (those that have already been treated) eliminate them from the dataframe.
  for(i in OLCount[,1]) {
    subcase<-subset(OrgList, OrgList[,3]==i)
    if (any(!is.na(DB[,7][which(DB[,6] %in% union(subcase[,1], subcase[,2]))]))) {
      for (j in union(subcase[,1], subcase[,2])) {
        if(any(!is.na(DB[,7][which(DB[,6]==j)]))) {
          subcase<-subcase[!(subcase[,1]==j | subcase[,2]==j),]
        }
      }
    }

    #After the previous operation, we verify that the subcase is not empty
    if(nrow(subcase)>0) {

      #Identify the cases with one dyad. These cases are k-places. Info is added to DB
      #k-place label is composed by the P + the number of the lowest place-number of the dyad + * +
      #the number of elements in the k-place + the number of sets in common + the value of k
      if(nrow(subcase)==1){
        DB[,7][which(DB[,6] %in% union(subcase[,1], subcase[,2]))]<-
          paste0("P", formatC(PrPl$PNumber[which(PrPl[,5]==subcase[1,1])], width=nchar(max(PrPl$PNumber)), flag="0"),
                 "*","(", as.factor(sum(PrPl$Cele[which(PrPl[,5] %in% union(subcase[,1], subcase[,2]))])),
                 "-", as.factor(length(unique(DB[,2][which(DB[,6] %in% union(subcase[,1], subcase[,2]))]))),
                 "-",as.factor(subcase[1,3]),"-",k,")")
        DB[,8][which(DB[,6] %in% union(subcase[,1], subcase[,2]))]<-
          paste(intersect(DB[,2][which(DB[,6]==subcase[1,1])], DB[,2][which(DB[,6]==subcase[1,2])]), collapse = ",")
      }

      #Identify the cases composed by more than one dyad. In these cases we can find k-places
      #identified in one o more dyads. We identify the string of sets in common for each dyad
      else {
        for(j in 1:length(subcase[,1])) {
          subcase$Inter[[j]]<-paste(intersect(DB[,2][which(DB[,6]==subcase[j,1])], DB[,2][which(DB[,6]==subcase[j,2])]), collapse = ",")
        }

        Int<-as.data.frame(unlist(unique(subcase[,6])))
        for(j in Int[,1]) {
          Int$Count[which(Int[,1]==j)]<-length(which(subcase[,6]==j))
        }
        #We prioritize the cases of dyads with the same string of shared sets in common.
        Int<-Int[order(rev(Int[,2]), Int[,1]),]

        for(j in c(1:length(Int[,1]))) {
          subcase2<-subset(subcase, subcase[,6]==Int[j,1])
          if (length(subcase2[,1])>1) {
            #Check if is an ambiguous case. An ambiguous case is a subset of places where the dyads
            #can be a kplace but not all the subset. For example, for k=1, P1 (AB) can be kplace
            #with P2 (ABC) or P3 (ABD), but the kplace formed by the three places is not true
            #because P2 and P3 have more than k different sets.
            if (c(length(unique(DB[,2][which(DB[,6] %in% union(subcase2[,1], subcase2[,2]))]))-subcase2[1,3])<=k) {
              DB[which(DB[,6] %in% union(subcase2[,1], subcase2[,2])),7]<-
                paste0("P",
                       formatC(PrPl$PNumber[which(PrPl[,5]==subcase2[1,1])], width=nchar(max(PrPl$PNumber)), flag="0"),
                       "*","(", as.factor(sum(PrPl$Cele[which(PrPl[,5] %in% union(subcase2[,1], subcase2[,2]))])),
                       "-", as.factor(length(unique(DB[,2][which(DB[,6] %in% union(subcase2[,1], subcase2[,2]))]))),
                       "-",as.factor(subcase2[1,3]),"-",k,")")
              DB[which(DB[,6] %in% union(subcase2[,1], subcase2[,2])),8]<-subcase2[1,6]
            }
            else {
              message("The places ", paste0(union(subcase2[,1], subcase2[,2]), collapse = " "), " are an ambiguous case")
            }
          }
          else {
            if (all(is.na(DB[,7][which(DB[,6] %in% union(subcase2[,1], subcase2[,2]))]))) {
              DB[,7][which(DB[,6] %in% union(subcase2[,1], subcase2[,2]))]<-
                paste0("P", formatC(PrPl$PNumber[which(PrPl[,5]==subcase2[1,1])], width=nchar(max(PrPl$PNumber)), flag="0"),
                       "*","(", as.factor(sum(PrPl$Cele[which(PrPl[,5] %in% union(subcase2[,1], subcase2[,2]))])),
                       "-", as.factor(length(unique(DB[,2][which(DB[,6] %in% union(subcase2[,1], subcase2[,2]))]))),
                       "-",as.factor(subcase2[1,3]),"-",k,")")
              DB[,8][which(DB[,6] %in% union(subcase2[,1], subcase2[,2]))]<-unlist(subcase2[1,6])
            } else { #When the k-place with the same number of sets exists
              subcase3<-subset(subcase, subcase[,1] %in% unlist(subcase2[,c(1,2)]) | subcase[,2] %in% unlist(subcase2[,c(1,2)]))
              #Remove data of k-places identified before
              DB[which(DB[,6] %in% unique(unlist(subcase3[,c(1,2)]))),7]<-NA
              DB[which(DB[,6] %in% unique(unlist(subcase3[,c(1,2)]))),8]<-NA
              message("The places ", paste0(sort(unique(unlist(subcase3[,c(1,2)]))), collapse = " "), " are an ambiguous case")
            }
          }
        }
      }
    }
  }

  #At this point, all k-places are identified
  #We create a dataframe with the kplaces, the places included in each one, and the common sets.
  KPlaList<-unique(DB[,c(7,6,8)][!is.na(DB[,7]),])

  #We modify the dataframe PrPl. First we change the first column, ProtoPlaces PP, in character
  #Then we calcule the new data for k-places.
  PrPl[,1]<-sapply(PrPl[,1], as.character)
  #Change the places merged by the k-places
  PrPl[,5] [which(PrPl[,5] %in% KPlaList[,2])]<-KPlaList[,1]
  #Recalculate number of elements, number of sets and protoplaces (detail places, PrPl[,1]) for k-places
  for (i in unique(KPlaList[,1])) {
    PrPl[,2] [which(PrPl[,5] == i)]<-length(unique(DB[,1][which(DB[,7]==i)]))
    PrPl[,3] [which(PrPl[,5] == i)]<-length(unique(DB[,2][which(DB[,7]==i)]))
    PrPl[,1] [which(PrPl[,5] == i)]<-
      paste0("{",paste(unique(DB[,1][which(DB[,7]==i)]), collapse = ";"),"}" ," - ","{" , paste(unique(DB[,2][which(DB[,7]==i)]), collapse = ";"),"}")
  }

  #Message with information about the number of places and k-places identified
  message("A total of ",length(unique(DB[,6][is.na(DB[,7])]))," places and ",
          length(unique(KPlaList[,1])) ," k-places have been identified")

  #Sort data in KPlaList by first and second col
  KPlaList<-KPlaList[order(KPlaList[,1], KPlaList[,2]),]

  #We complete DB$k_places with places and we create the edgelist
  DB[,7][is.na(DB[,7])]<-DB[,6][is.na(DB[,7])]
  KEdList<-unique(DB[,c(7,2)])

  #Objets for output are created
  KPlacesResult<-list()
  KPlacesResult$data<-DB[,c(1,2,7)]
  KPlacesData<-unique(PrPl[,c(5,2,3,1)])
  names(KPlacesData)[1]<-"PlaceLabel"
  names(KPlacesData)[2]<-"NbElements"
  names(KPlacesData)[3]<-"NbSets"
  names(KPlacesData)[4]<-"PlaceDetail"
  KPlacesResult$KPlacesData<-KPlacesData
  KPlacesResult$kPlaces<-KPlaList
  KPlacesResult$Edgelist<-KEdList

  return(KPlacesResult)
}
