#' @title places
#' @description Execute a structural equivalence analysis in bipartite graphs
#' @usage places(data, col.elements, col.sets)
#' @param data The input data frame. Data must be in a two-mode edgelist structure
#' @param col.elements The name of the column of elements
#' @param col.sets The name of the column of sets
#' @format The input format must be a data frame in a two-mode edgelist structure. \cr
#' Data frame can contain other variables, relevant columns will be selected in \emph{Arguments} parameters. Missing values, empty cells and repeated pairs (Element-Set) will be deleted for analysis.
#' @details A place is a structurally equivalent position defined by one or more sets, containing one or more elements. Then, two or more elements share a place (so, are structurally equivalent) if they are affiliated to the same sets. \cr
#' A network of places is defined as a collection of places linked by sharing sets. A pair of places are linked if they are defined by at least one set in common. A network of places, as a blockmodel image, is a network of structurally equivalent positions.
#' @return The output is a list with three data frames:
#' @return \itemize{\item The original two columns data frame and the column "Places" with places labels
#' \item A data frame containing information about places.\cr
#' "PlaceNumber" contains the number of the place, ordered from the highest to the lowest number of sets \cr
#' "PlaceLabel" contains places labels. Labels start with \emph{P}, followed by the place number, the number of elements in place and the number of sets defining place \cr
#' "NbElements" contains the number of elements in the place \cr
#' "NbSets" contains the number of sets defining the place \cr
#' "PlaceDetail" contains the name of all the elements in the place and all the sets defining the place
#' \item The network of places in a two-mode edgelist format }
#' @author Delio Lucena Piquero (\href{http://lereps.sciencespo-toulouse.fr/}{LEREPS} - University of Toulouse) <delio.lucena@ut-capitole.fr>
#' @references Pizarro, N. (2007). Structural identity and equivalence of individuals in social networks: Beyond duality. International Sociology, 22(6), 767-792 \cr
#' Pizarro, N. (2000).  Regularidad relacional, redes de lugares y reproducción social. Política y Sociedad, Vol 33, 167-198 \cr
#' Pizarro, N. (1999).  Appartenances, places et réseaux de places: la reproduction des processus sociaux et la génération d'un espace homogène pour la définition des structures sociales. Sociologie et Societés, Vol 31(1), 143-161
#' @examples # Load example dataset
#' data(fooA)
#'
#' # Compute places
#' Result <- places(data = fooA, col.elements = "Actors", col.sets = "Groups")
#' Result <- places(fooA, "Actors", "Groups")
#' @importFrom stats aggregate na.omit
#' @export

places<-function(data, col.elements, col.sets) {
  #Subset data into two variable data frame, elements and sets
  pos.x<-which(colnames(data)==col.elements)
  pos.y<-which(colnames(data)==col.sets)
  DB<-data[,c(pos.x, pos.y)]
  #Remove empty cells and NAs. Repport the information
  message("Cleaning data ... rows with empty cells and NAs will be removed")
  DB[DB==""]<-NA
  frm<-nrow(DB)
  DB<-na.omit(DB)
  message("Rows removed: ", c(frm - nrow(DB)))

  message("Cleaning data ... duplicate rows will be removed")
  frm<-nrow(DB)
  DB<-unique(DB)
  message("Duplicate rows removed: ", c(frm - nrow(DB)))

  #Identify the number of elements and sets
  message("There are ",length(unique(DB[,1]))," elements and ",length(unique(DB[,2]))," sets")

  message("Working ...")

  #Rename column names to "Elem" and "Set"
  names(DB)[1]<-"Elem"
  names(DB)[2]<-"Set"

  #Identify the number of Sets by Element
  #Calculate the frequency of each Element in edgelist
  count_i<-as.data.frame(table(DB[,1]))
  #Add the result to data frame as new variable called "Freq"
  DB$count_i<-count_i[match(DB$Elem, count_i$Var1),"Freq"]

  #Create the variables "count_g", "count_dif_i", "ProtoPlaces" and "Places"
  DB$count_g<-NA
  DB$count_dif_i<-NA
  DB$ProtoPlaces<-NA
  DB$Places<-NA

  #Order the data by some variables
  DB<-DB[order(-DB[,3], DB[,1], DB[,2]),]

  #Create "count_i2". This object record all the differents number of Sets by place
  #Each value indicates that there is at least one place with that number of sets
  #This vector drives the loop
  count_i2<-unique(DB$count_i)

  for (i in count_i2) {
    subcase<-subset(DB, count_i==i)
    El<-unique(subcase[,1])
    #The values of the variable "Set" are aggregates as list
    subcase2<-aggregate(Set~Elem, data=subcase, FUN=list)
    #Information of variable "count_g" added
    subcase$count_g<-length(unlist(subcase2[1,2]))
    #Values of "Set" are reclassified in character to be aggregated again
    subcase2$Set<-sapply(subcase2$Set, paste, collapse=";")
    #Aggregate again in order to have all the elements together
    subcase3<-aggregate(Elem~Set, data=subcase2, FUN=list)
    #Count the number of elements per place
    subcase3$count_dif_i<-sapply(subcase3$Elem, function(x) length(unlist(x)))
    #The values of "Elem" are reclassified in character
    subcase3$Elem<-sapply(subcase3$Elem, paste, collapse=";")
    #Places are identified. Variables "Pelem" and "Pset" are added and "Protoplaces" is created
    subcase3$Pelem<-sapply(subcase3$Elem, paste, collapse=";")
    subcase3$Pset<-subcase3$Set
    subcase3$ProtoPlace<-paste0("{", subcase3$Pelem,"}" ," - ","{", subcase3$Pset,"}")
    #Variable "Pset" is created in subcase2
    subcase2$Pset<-subcase2$Set
    #Information is added to subcase. First the "Pset" from subcase2 and then "Pelem",
    #"count_dif_i" and "Protoplace" from subcase3 through "Pset"
    subcase<-merge(subcase, subcase2[,c("Elem","Pset")], by="Elem")
    subcase$count_dif_i<-NULL
    subcase<-merge(subcase, subcase3[,c("Pset","Pelem","count_dif_i" ,"ProtoPlace")], by="Pset")

    #Information about "Protoplaces" and "count_g" is added to DB
    for (j in El) {
      DB[,"ProtoPlaces"][which(DB$Elem == j)]<-subcase[,"ProtoPlace"][which(subcase$Elem == j)]
      DB[,"count_dif_i"][which(DB$Elem == j)]<-subcase[,"count_dif_i"][which(subcase$Elem == j)]
      DB[,"count_g"][which(DB$Elem == j)]<-subcase[,"count_g"][which(subcase$Elem == j)]
    }
  }

  #To add places' names a new loop is created for the suite
  #All places are listed in a new data frame. Places will be added to BD later.
  PP<-unique(DB$ProtoPlaces)
  PrPl<-as.data.frame(PP)
  PrPl$Cele<-NA
  PrPl$Cset<-NA

  for(k in PrPl$PP) {
    PrPl[,"Cele"][which(PrPl$PP==k)]<-unique(DB[,"count_dif_i"][which(DB$ProtoPlaces==k)])
    PrPl[,"Cset"][which(PrPl$PP==k)]<-unique(DB[,"count_g"][which(DB$ProtoPlaces==k)])
  }

  #Data frame is ordered by number of sets and by number of elements
  PrPl<-PrPl[order(-PrPl[,3], -PrPl[,2]),]

  #New variables are added and the label of each place is created
  PrPl$PNumber<-(1:length(PrPl$PP))
  PrPl$PName<-NA
  #For places' label we use le letter "P" followed by the number of the place and zeros on
  #the left, the number of elements and the number of sets
  for(k in PrPl$PP) {
    PrPl$PName<-paste0("P", formatC(PrPl$PNumber, width=nchar(max(PrPl$PNumber)), flag="0"),
                       "(",as.factor(PrPl$Cele),"-", PrPl$Cset,")")
  }

  #INformation of places is added to DB
  for(l in DB$ProtoPlaces) {
    DB[,"Places"][which(DB$ProtoPlaces==l)]<-PrPl[,"PName"][which(PrPl$PP==l)]
  }

  #Variable "count_g" is deleted because is the same than "count_i"
  DB$count_g<-NULL

  #We create the edgelist
  #DB is ordered by number of sets, number of elements and places
  DB<-DB[order(DB[,6]),]
  EdList<-unique(DB[,c(6,2)])

  #Objects for output are created
  PlacesResult<-list()
  PlacesResult$data<-DB[,c(1,2,6)]
  PlacesData<-PrPl[,c(4,5,2,3,1)]
  names(PlacesData)[1]<-"PlaceNumber"
  names(PlacesData)[2]<-"PlaceLabel"
  names(PlacesData)[3]<-"NbElements"
  names(PlacesData)[4]<-"NbSets"
  names(PlacesData)[5]<-"PlaceDetail"
  PlacesResult$PlacesData<-PlacesData
  PlacesResult$Edgelist<-EdList

  message("A total of ",nrow(PrPl)," places have been identified")

  return(PlacesResult)
}
